
class TMCLException(Exception):
    pass

class ObjectClosedException(TMCLException):
    pass

class TMCLBadResponse(TMCLException):
    pass
class MultiAxisException(TMCLException):
    pass

class GPIOException(TMCLException):
    pass
class PinException(GPIOException, ValueError):
    pass


from typing import TYPE_CHECKING, Tuple
if TYPE_CHECKING:
    from .tmcl_board import TMCLBoard

from .exceptions import PinException, ObjectClosedException

# Pin names:
# Use either an int = (10*bank number) + pin number
# eg. bank 0 pin 4 = 4, bank 2 pin 1 = 21
# Or str of the form in the tmcm manual (IN_3, OUT_7, AIN_1, etc.)


class TMCLGPIO:
    
    def __init__(self, parent_board: TMCLBoard):
        # Reference can be used for coordinating movements and to ensure board object isn't deleted before gpio and motors are done.
        self.parent_board = parent_board
    
    def assert_parent_open(self):
        if self.parent_board.is_closed:
            raise ObjectClosedException()

    def pin_lookup(self, pin:"int | str") -> Tuple[int, int]:
        """Returns bank, pin_number."""
        self.assert_parent_open()
        if isinstance(pin, int):
            bank_num = pin // 10
            pin_num = pin % 10
            if bank_num not in (0, 1, 2):
                raise PinException(f"{pin} is not in bank 0, 1 or 2.")

        elif isinstance(pin, str):
            split_str = pin.split("_")
            if len(split_str) != 2:
                raise PinException(f"{pin} does not contain exactly one underscore.")
            bank_str = split_str[0]
            pin_str = split_str[1]

            try:
                pin_num = int(pin_str)
            except ValueError:
                raise PinException(f"{pin} does not end in valid base 10 number.")

            if bank_str in ("IN", "AIN"):
                bank_num = 0
            elif bank_str == "OUT":
                bank_num = 2
            else:
                raise PinException(f"{pin} does not begin with AIN, IN, or OUT")
        
        else:
            raise ValueError(f"{pin} is not an int or a str.")
            
        
        return bank_num, pin_num


    def digital_write(self, pin:"int | str", value:bool):
        """Set the specified pin to the specified value."""
        self.assert_parent_open()
        bank_num, pin_num = self.pin_lookup(pin)
        if bank_num != 2:
            raise PinException(f"{pin} is an input pin.")
        self.parent_board.command("Set IO", pin_num, 2, int(bool(value))) # Output pins are bank 2
    
    def digital_read(self, pin:int) -> bool:
        """Get the value of the specified digital pin."""
        self.assert_parent_open()
        bank_num, pin_num = self.pin_lookup(pin)
        return bool(self.parent_board.command("Get IO", pin_num, bank_num, 0)) # Digital output is on bank 2
    
    def analog_read(self, pin:int) -> int:
        """Get the value of the specified analog pin."""
        self.assert_parent_open()
        bank_num, pin_num = self.pin_lookup(pin)
        if bank_num == 0 and pin_num not in (0, 4): # Digital input pin
            raise PinException(f"{pin} is a digital input pin.")
        elif bank_num == 2:
            raise PinException(f"{pin} is a digital output pin.")
        return self.parent_board.command("Get IO", pin_num, 1, 0) # Analog input is bank 1


import serial.tools.list_ports
import serial.tools.list_ports_common
from serial import Serial
import platform

from .tmcl_board import TMCLBoard


def port_name(port: serial.tools.list_ports_common.ListPortInfo) -> str:
    """Takes a port object and returns the string name that can be used to open it as a serial connection."""
    if platform.system() == "Windows":
        port_name = port.name
    elif platform.system() == "Linux":
        port_name = "/dev/" + port.name
    else:
        raise Exception
    return port_name

def get_trinamics() -> "dict[int:TMCLBoard]":
    """Return a dictionary of Trinamic board serial addresses and associated objects."""
    boards = {}

    ports = serial.tools.list_ports.comports()

    for p in ports:
        if "VID:PID=2A3C:0100" in p.hwid or "VID:PID=16D0:0653" in p.hwid: # If it is a Trinamic TMCM-1110/3110/6110 Board
            board = TMCLBoard(port_name(p)) # Create board object
            address = board.get_param("Serial Address") # Get Address

            boards[address] = board

    return boards



from __future__ import annotations

from typing import Any
import json


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .tmcl_board import TMCLBoard
from .exceptions import ObjectClosedException

class TMCLMotion:
    """
        Factory class for TrinamicMotion.
        The details of initialization and conversion can vary between boards.
        The subclass used is specified by the "Motion Type" parameter in the trinamic section of the json file.
    """

    def __new__(cls, parent_board: TMCLBoard, axis_number:int, config:"dict[str:Any] | str"):
        if isinstance(config, str):
            # Open JSON config if specified
            with open(config) as config_file:
                config = json.load(config_file)
        else:
            # Assuming config is a dict
            config = config

        try:
            motion_type = config["Trinamic"]["Motion Type"]
        except KeyError:
            motion_type = None

        if motion_type == "External Driver":
            return super().__new__(TrinamicExternal)
        else:
            return super().__new__(TrinamicDefault)

    def __init__(self, parent_board: TMCLBoard, axis_number:int, config:"dict[str:Any] | str"):
        # Reference can be used for coordinating movements and to ensure board object isn't deleted before gpio and motors are done.
        self.parent_board: TMCLBoard = parent_board
        self.axis_number = axis_number

        # Load config file if specified
        if type(config) == str:
            # Open JSON config if specified
            with open(config) as config_file:
                self.config = json.load(config_file)
        else:
            # Assuming config is a dict
            self.config = config

        if "Left Limit Switch Disable" not in self.config["Trinamic"]:
            self.set_param("Left Limit Switch Disable", 0)
        if "Right Limit Switch Disable" not in self.config["Trinamic"]:
            self.set_param("Right Limit Switch Disable", 0)
        if "Reference Search Coarse Speed" not in self.config["Trinamic"]:
            self.set_param("Reference Search Coarse Speed", self.get_param("Maximum Speed"))
        if "Reference Search Coarse Speed" not in self.config["Trinamic"]:
            self.set_param("Reference Search Fine Speed", 10)

        # Initialize based on config file
        for key, value in self.config["Trinamic"].items():
            if key != "Motion Type":
                self.set_param(key, value)

        # Initialize General config. Must be after Trinamic section because of things like divisors and microstep settings.
        super().__init__(self.config["General"])

    def assert_parent_open(self):
        if self.parent_board.is_closed:
            raise ObjectClosedException()

    # Axis Parameters
    def set_param(self, parameter:str, value:int):
        """Set the value of the specified register."""
        self.assert_parent_open()
        self.parent_board.command("Set Axis", parameter, self.axis_number, value)

    def get_param(self, parameter:str) -> int:
        """Return the value of the specified register."""
        self.assert_parent_open()
        return self.parent_board.command("Get Axis", parameter, self.axis_number, 0)

    # Movement
    def move_relative(self, distance:float):
        """Move motor a relative distance."""
        self.assert_parent_open()
        self.move_steps_relative(round(self.units_to_steps(distance)))

    def move_absolute(self, position:float):
        """Move motor to an absolute position based on current 0 position."""
        self.assert_parent_open()
        self.move_steps_absolute(round(self.units_to_steps(position)))

    def move_velocity(self, velocity:float):
        """Move indefinitely at the specified velocity."""
        self.assert_parent_open()
        self.move_vinternal_velocity(round(self.vpps_to_vint(self.units_to_steps(velocity))))

    def move_steps_relative(self, steps:int):
        """Move motor a specific number of steps."""
        self.assert_parent_open()
        self.parent_board.command("Move", "Relative", self.axis_number, steps)

    def move_steps_absolute(self, steps:int):
        """Move motor to an absolute position based on current 0 position."""
        self.assert_parent_open()
        self.parent_board.command("Move", "Absolute", self.axis_number, steps)

    def move_vinternal_velocity(self, velocity:int):
        self.assert_parent_open()
        if(velocity < 0):
            self.parent_board.command("Rotate Left", 0, self.axis_number, abs(velocity))
        else:
            self.parent_board.command("Rotate Right", 0, self.axis_number, velocity)

    def stop(self):
        """Stop the motor."""
        self.assert_parent_open()
        self.parent_board.command("Stop", 0, self.axis_number, 0)

    def home(self, to_right=False):
        """Home the motor. If to_right is true, it will home to the positive direction. Else, it will home to the negative direction."""
        self.assert_parent_open()
        self.set_param("Reference Search Mode", 65 if to_right else 1)
        self.parent_board.command("Reference Search", "Start", self.axis_number, 0)

    ##Getting position
    def get_position(self) -> float:
        """Get position of motor."""
        self.assert_parent_open()
        return self.steps_to_units(self.get_steps())

    def get_steps(self) -> int:
        """Get position of motor in steps"""
        self.assert_parent_open()
        return self.get_param("Actual Position")

    def set_speed(self, speed:float):
        """Set the speed of the stage."""
        self.assert_parent_open()
        self.set_param("Maximum Speed", self.convert_speed(speed))
    
    def get_speed(self) -> float:
        self.assert_parent_open()
        return self.steps_to_units(self.vint_to_vpps(self.get_param("Maximum Speed")))

    def set_acceleration(self, acceleration:float):
        """Set the acceleration of the stage."""
        self.assert_parent_open()
        self.set_param("Maximum Acceleration", self.convert_acceleration(acceleration))
    
    def get_acceleration(self) -> float:
        self.assert_parent_open()
        return self.steps_to_units(self.aint_to_apps(self.get_param("Maximum Acceleration")))

    ##Getting State
    def is_moving(self) -> bool:
        """
        Return if motor is moving. This works regardless of whether the motor is in position mode, velocity mode, or a reference search.

        Important: since this gets the current velocity of the motor, this will return False for a millisecond or two after starting a movement.
        Use a short delay first if waiting for the motor to arrive somewhere.

        The main other candidate we have used in the past is the position reached flag, but that doesn't go on if the motor hits a limit switch.
        The target velocity behaves similarly. If you have any suggestions, please message Gideon Rabson in slack.
        """
        self.assert_parent_open()
        return self.get_param("Actual Speed") != 0
    
    def is_homing(self) -> bool:
        """Return if the motor is homing."""
        self.assert_parent_open()
        return self.parent_board.command("Reference Search", "Status", self.axis_number, 0) != 0

    # Steps and units
    def units_to_steps(self, value:float) -> float:
        """Convert a value from stage units to steps."""
        raise NotImplementedError

    def steps_to_units(self, steps:float) -> float:
        """Convert a value from steps to stage units."""
        raise NotImplementedError

    # Internal units
    # Equations from firmware manual chapter 6.1
    def vint_to_vpps(self, Vint:float) -> float:
        self.assert_parent_open()
        return float(Vint) * (16 * (10**6)) / ((2**self.get_param("Pulse Divisor")) * 2048 * 32)

    def vpps_to_vint(self, Vpps:float) -> float:
        """Convert velocity from steps per second to internal units. Must be converted to int in [0..2047] before use."""
        self.assert_parent_open()
        return float(Vpps) * ((2**self.get_param("Pulse Divisor")) * 2048 * 32) / (16 * (10**6))

    def aint_to_apps(self, Aint:float) -> float:
        """Convert a value from steps to stage units."""
        self.assert_parent_open()
        return float(Aint) * ( (16 * (10**6)) ** 2) / (2 ** (self.get_param("Ramp Divisor") + self.get_param("Pulse Divisor") + 29))

    def apps_to_aint(self, Apps:float) -> float:
        """Convert acceleration from steps per second^2 to internal units. Must be converted to int in [0..2047] before use."""
        self.assert_parent_open()
        return float(Apps) * (2 ** (self.get_param("Ramp Divisor") + self.get_param("Pulse Divisor") + 29)) / ( (16 * (10**6)) ** 2)

    # Overflow prevention
    def convert_speed(self, speed:float) -> int:
        self.assert_parent_open()
        return max(1, min(round(self.vpps_to_vint(self.units_to_steps(speed))), 2047))

    def convert_acceleration(self, acceleration:float) -> int:
        self.assert_parent_open()
        return max(1, min(round(self.apps_to_aint(self.units_to_steps(acceleration))), 2047))
    
    def disable(self):
        raise NotImplementedError
    
    def enable(self):
        raise NotImplementedError

class TrinamicDefault(TMCLMotion):
    def __init__(self, parent_board: TMCLBoard, axis_number: int, config: "dict[str:Any] | str"):
        super().__init__(parent_board, axis_number, config)
        if "Chopper Off Time" not in self.config["Trinamic"]:
            self.config["Trinamic"]["Chopper Off Time"] = 5     # Seems to be the default, no reason to mess with it
        self.enable()

    def disable(self):
        """Cut power to motor. It is unclear whether current level 0 is actually 0 current, but I hope so."""
        self.assert_parent_open()
        self.set_param("Chopper Off Time", 0)
    
    def enable(self):
        """Only needed after disabling. Motor is automatically enabled on initialization."""
        self.assert_parent_open()
        self.set_param("Chopper Off Time", self.config["Trinamic"]["Chopper Off Time"])

    def units_to_steps(self, value:float) -> float:
        """Convert a value from stage units to steps."""
        self.assert_parent_open()
        return float(value) * (2**self.get_param("Microstep Resolution")) / self.config["General"]["distance per step"]

    def steps_to_units(self, steps:float) -> float:
        """Convert a value from steps to stage units."""
        self.assert_parent_open()
        return float(steps) / (2**self.get_param("Microstep Resolution")) * self.config["General"]["distance per step"]

class TrinamicExternal(TMCLMotion):
    def units_to_steps(self, value:float) -> float:
        """Convert a value from stage units to steps."""
        self.assert_parent_open()
        return float(value) / self.config["General"]["distance per step"]

    def steps_to_units(self, steps:float) -> float:
        """Convert a value from steps to stage units."""
        self.assert_parent_open()
        return float(steps) * self.config["General"]["distance per step"]


from typing import Any, List, Tuple

from serial import Serial
import struct
import json
import importlib.resources
import threading
import weakref
import time

from .exceptions import *
from .tmcl_motion import TMCLMotion
from .tmcl_gpio import TMCLGPIO



with importlib.resources.open_text("automationlib.trinamic", "TMCLCommands.json") as commandsFile:
    _DATA = json.load(commandsFile)


COMMANDS = _DATA["Commands"]
PARAMETERS = _DATA["Parameters"]
ERRORS = _DATA["Errors"]



MESSAGE_STRUCTURE=">BBBBiB" #byte,byte,byte,byte,integer,byte
REPLY_STRUCTURE=">BBBBiB" #byte,byte,byte,byte,integer,byte


# Fairly certain the occaisional weird error where it returns a status code of 0 are due to the system not being thread-safe.
# Adding a threading lock on writing to test.

class TMCLBoard():
    """Class to handle interaction with the trinamic control boards."""

    def __init__(self, portName):
        self.serial_port = Serial(portName, timeout=0.2)
        self.serial_lock = threading.Lock()

        self.close = weakref.finalize(self, self._close, self.serial_port)

    @staticmethod
    def _close(serial_port:Serial):
        serial_port.close()
    
    @property
    def is_closed(self):
        return not self.close.alive
    
    def assert_open(self):
        if self.is_closed():
            raise ObjectClosedException()
    
    def get_motion(self, axis: int, config:"dict[str:Any] | str") -> TMCLMotion:
        """Get a motion object corresponding to a specified axis of the board."""
        self.assert_open()
        return TMCLMotion(self, axis, config)
    
    def get_gpio(self):
        """Get the gpio object for the gpio pins on the board."""
        self.assert_open()
        return TMCLGPIO(self)
    
    def set_param(self, parameter:"int | str", value:int):
        """Set global parameter."""
        self.assert_open()
        self.command("Set Global", parameter, 0, value)
    def get_param(self, parameter:"int | str") -> int:
        """Get global parameter."""
        self.assert_open()
        return self.command("Get Global", parameter, 0, 0)

    def set_coordinate(self, coordinate_number:int, axis:TMCLMotion, position:float):
        """Set the value of the coordinate for the specified axis."""
        self.assert_open()
        self.command("Set Coordinate", coordinate_number, axis.axis_number, round(axis.units_to_steps(position)))
    def get_coordinate(self, coordinate_number:int, axis:TMCLMotion):
        """Get the value of the coordinate for the specified axis"""
        self.assert_open()
        return axis.steps_to_units(self.command("Get Coordinate", coordinate_number, axis.axis_number, 0))
    
    def move_coordinate(self, coordinate_number:int, axes:List[TMCLMotion]):
        """Move multiple axes together to the coordinate specified. Behaves strangely for axes with low acceleration."""
        self.assert_open()
        if len(axes) > 3:
            raise MultiAxisException(f"Move coordinate can only move 3 axes with interpolation. {len(axes)} were specified.")
        
        axis_bitmask = 0b01000000 # 6th bit specifies move together
        for axis in axes:
            if not axis.parent_board is self:
                raise MultiAxisException(f"Axis {axis} is not connected to this control board.")
            axis_bitmask = axis_bitmask | (1 << axis.axis_number) # Set the bit corresponding to the axis number of the axes

        self.command("Move", "Coordinate", axis_bitmask, coordinate_number)
    
    def move_absolute(self, positions:List[Tuple[TMCLMotion, float]], coordinate_number:int=1):
        """Move the specified axes to the specified positions."""
        self.assert_open()
        if len(positions) > 3:
            raise MultiAxisException(f"Move coordinate can only move 3 axes with interpolation. {len(positions)} were specified.")
        
        for position in positions:
            self.set_coordinate(coordinate_number, position[0], position[1])
        self.move_coordinate(coordinate_number, [position[0] for position in positions])
    
    def command(self, command:"int | str", type_:"int | str", bank:int, value:int) -> int:
        """Write the specified command to the Trinamic board and return the board's response."""
        self.assert_open()

        def calc_checksum(message_without_checksum:struct) -> int:
            """Calculate the checksum of the message"""
            checksum = 0
            for i in message_without_checksum:
                checksum += int(i)

            # Gets the least significant byte
            checksum = checksum & 255
            return checksum

        # Moved these functions inside since they should NEVER be called by themselves.
        def write(self, command:int, type_:int, bank:int, value:int):
            # To calculate checksum, format message without checksum
            checksum = calc_checksum(struct.pack(MESSAGE_STRUCTURE[:-1], 1, command, type_, bank, value))

            # Create and send message
            message = struct.pack(MESSAGE_STRUCTURE, 1, command, type_, bank, value, checksum)
            self.serial_port.write(message)
        
        def reply(self) -> int:
            replyBytes = self.serial_port.read(9)
            if len(replyBytes) < 9:
                raise TMCLBadResponse("Board timed out! Check connection.")
            reply = struct.unpack(REPLY_STRUCTURE,replyBytes)

            status_code = reply[2]
            value = reply[4]

            # Raise exception if it doesn't give a happy reply
            if status_code not in (100, 101):
                try:
                    error_message = ERRORS[str(status_code)]
                except:
                    error_message = f"Unknown error code: {status_code}"

                raise TMCLBadResponse(error_message)
            else:
                return value
        

        with self.serial_lock:
            # If the command or type are strings, try looking them up in the json file.
            if type(command) == int:
                command_num = command
            else:
                command_num = COMMANDS[command]

            if type(type_) == int:
                type_num = type_
            else:
                type_num = PARAMETERS[type_]

            write(self, command_num, type_num, bank, value)
            return reply(self)
    
    def reset_serial(self):
        """Used to reset the serial connection in the case that something went wrong (eg. a thread was aborted while sending a command.)."""
        self.serial_port.reset_output_buffer()
        time.sleep(0.002) # Give the trinamic board a moment to make sure everything's processed.
        self.serial_port.reset_input_buffer()

        # Release the lock if it's stuck.
        if self.serial_lock.locked():
            self.serial_lock.release()
        
